// (c) 2014, SAS Institute Inc.

jQuery.sap.declare("sas.hc.m.BannerHeading");
jQuery.sap.require("sas.hc.m.Heading");

jQuery.sap.require("sap.ui.core.Core");
jQuery.sap.require("sas.hc.ui.core.Core");
jQuery.sap.require("sas.hc.ui.core.util.ServicesUtil");

/**
 * Is used for creating a banner heading within a page. Maps directly to the H1, H2, etc. elements in HTML
 * Banner headings are unique in that they can have two parts: the heading and a task text.  The latter is optional.
 * Another feature is that the heading visibility can be controlled with the showText property.   E.g. when the browser
 * narrows enough and there are both a heading and a task, the banner suppresses the heading and shows only the task.
 *
 * This is only for use in WIP deployments.  For Viya, use the BannerV component.
 *
 * @class
 * @name sas.hc.m.BannerHeading
 * @extends sas.hc.m.Heading
 */
sas.hc.m.Heading.extend("sas.hc.m.BannerHeading", /** @lends sas.hc.m.BannerHeading.prototype */ {
    metadata: {

        library: "sas.hc.m",

        properties: {
            /** This an optional property that describes the current module or task. */
            "task": {type: "string", group: "Misc", defaultValue: ""},

            /** If false, only task/subText is displayed.  Only false if window size is not big enough for everything */
            "showText": {type: "boolean", group: "Misc", defaultValue: "true"},

            /** subText should not be defined w/o also a "task" */
            "subText": {type: "string", group: "Misc", defaultValue: ""}
        }
    },

    rb : sap.ui.getCore().getLibraryResourceBundle("sas.hc.m"),

    renderer: "sas.hc.m.BannerHeadingRenderer"

});

/**
 * Initialize the component.
 *
 * @override
 *
 */
sas.hc.m.BannerHeading.prototype.init = function () {


    var s;

    // Find the name/verb/subText from the config.  Given an app switcher, use it to find the app name.
    if (sas && sas.config) {
        // No app switcher, use the config.
        s =  sas.config.getAppName();
    }

    this.setProperty("text", s);

};


/**
 * Dynamically add a tooltip if the title is truncated.
 *
 * @param {*}
 * @public
 * @name sas.hc.m.Heading#onAfterRendering
 * @function
 */
sas.hc.m.BannerHeading.prototype.onAfterRendering = function() {

    var me = this,
        bFirstHeaderTruncated,
        sTooltip;

    this.$().css("overflow", "hidden");
    this.$().css("text-overflow", "ellipsis");
    this.$().css("text-align", "center");

    var oDomRef = this.getDomRef();
    if ( oDomRef ) {

        // This mouseenter operates on the outer div.
        $(oDomRef).bind('mouseenter', function(){
            var $m = $(this);  // This is the outer div.
            var $firstHeader = $("h" + me.getLevel(), $m)[0];
            bFirstHeaderTruncated = false;

            // If we don't have a tooltip (set by consumer) check to see if the text is truncated.
            // If so, then create a tooltip on the fly of the whole text.
            // There are 3 strings that can get truncated: the text by itself, the task by itself,
            // and the task + subtext.  In all cases, create a tooltip of everything that's available.
            if (me && (!me.getTitle() || me.getTitle().length === 0)) {

                // First create the tooltip that will be used in all cases.
                sTooltip = me.getText();
                if (me.getTask() && me.getTask().length > 0 ) {

                    // Note that sub text can't exist without a task.  Or that's the rule at least.
                    if (me.getSubText() && me.getSubText().length > 0) {
                        sTooltip = me.rb.getText("bannerHeading.textAndTaskAndSubText.fmt", [me.getText(), me.getTask(), me.getSubText()]);
                    } else {
                        sTooltip = me.rb.getText("bannerHeading.textAndTask.fmt", [me.getText(), me.getTask()]);
                    }
                }

                // If the text is hidden, this is a form of truncation.
                if ( !me.getShowText() || $firstHeader.offsetWidth < $firstHeader.scrollWidth) {
                    bFirstHeaderTruncated = true;
                }

                if (bFirstHeaderTruncated) {
                    $m.attr('title', sTooltip);
                    me._titleSetInternally = true;
                } else
                if ($m.attr("title") && me._titleSetInternally) {

                    // Nothing truncated, remove any title
                    $m.removeAttr("title");
                    me._titleSetInternally = false;
                }
            }
        });
    }
};

/**
 * Return the whole string (text and task) as a formatted string.
 *
 * @return string
 */
sas.hc.m.BannerHeading.prototype.getFormattedString = function() {

    var s;
    if (this.getText() && this.getShowText()) {
        s = this.getText();
    }

    // Get the task text.
    if (this.getTask()) {
        if (this.getShowText()) {
            s = this.rb.getText("bannerHeading.textAndTask.fmt", [this.getText(), this.getTask()]);
        } else {
            s = this.getTask();
        }
    }
    return s;
};

/**
 * This is the app sub text.
 *
 * @return {*}
 */
sas.hc.m.BannerHeading.prototype.getSubText = function() {
    //  Unless app explicitly overrides appname via setText we will
    // look up the appName via prepopulated config object
    var sSubText = this.getProperty("subText");
    var sReturnValue = "";
    if (sSubText !== "") {
        sReturnValue = sSubText;
    } else if (sas.applicationSwitcher && sas.config && sas.hc.ui.core.util.ServicesUtil.getServicePlatform() === "PETRICHOR") {
        var sAppId = sas.config.getAppId();
        var fCompare = function(oApp) {
            return oApp.appId === sAppId;
        };

        var oApplication = sas.applicationSwitcher.findApplication(fCompare);

        if (oApplication && oApplication.label && !oApplication.isHome) {
            var sHtmlText = $("<div></div>").html(oApplication.label).text();
            if (sHtmlText !== this.getText()) {
                sReturnValue = sHtmlText;
            } else {
                sReturnValue = "";
            }
        }
    }

    return sReturnValue;
};
