// (c) 2018, SAS Institute Inc.
sap.ui.define([
    "jquery.sap.global",
    "sap/ui/core/Core",
    "./WelcomeEvent",
    "sas/hc/m/Dialog",
    "sas/hc/ui/core/Core"
], function(jQuery, sapCore, WelcomeEvent, Dialog, Core) {
    "use strict";

    var Animation = {
        Direction: { // lower-case because they double as CSS style class names
            Forward: "forward",
            Back: "back",
            Skip: "skip"
        },
        Action: { // lower-case because they double as CSS style class names
            Enter: "enter",
            Exit: "exit"
        }
    };

    var ANIMATION_STYLE_CLASS = "sasWelcomeAnimation";

    var ALL_ANIMATION_STYLE_CLASSES = [
        ANIMATION_STYLE_CLASS,
        Animation.Direction.Forward,
        Animation.Direction.Back,
        Animation.Direction.Skip,
        Animation.Action.Enter,
        Animation.Action.Exit
    ].join(" ");


    var PAGE_VIEW_NAMES = [
        "sas.hc.m.welcome.WelcomeIntro",
        "sas.hc.m.welcome.WelcomeProfile",
        "sas.hc.m.welcome.WelcomeThemes",
        "sas.hc.m.welcome.WelcomeOutro"
    ];

    var MAX_ANIMATION_DURATION_MS = 10000; // 10s

    /**
     * Provides the dialog containing the Welcome Experience.
     * @name sas.hc.m.welcome.WelcomeDialog
     * @extends sas.hc.m.Dialog
     * @author madeze
     */
    return Dialog.extend("sas.hc.m.welcome.WelcomeDialog", /** @lends sas.hc.m.welcome.WelcomeDialog.prototype */ {

        init: function() {
            // logger
            this._oLogger = sas.log.getLogger("sas.hc.m.welcome.WelcomeDialog" /*, sas.log.Level.DEBUG*/);
            this._oLogger.debug("init()");

            if (Dialog.prototype.init) {
                Dialog.prototype.init.apply(this, arguments);
            }

            // configure general UI info
            this.setContentWidth("28rem");
            this.setContentHeight("32rem");
            this.setShowHeader(false);
            this.setShowClose(false);
            this.addStyleClass("sasWelcomeDialog");

            // subscribe to welcome events
            sas.eventBus.subscribe(WelcomeEvent.EventBusChannelId, WelcomeEvent.Previous, this._onPrevious, this);
            sas.eventBus.subscribe(WelcomeEvent.EventBusChannelId, WelcomeEvent.Next, this._onNext, this);
            sas.eventBus.subscribe(WelcomeEvent.EventBusChannelId, WelcomeEvent.Skip, this._onSkip, this);
            sas.eventBus.subscribe(WelcomeEvent.EventBusChannelId, WelcomeEvent.AnimationStart, this._onAnimationStart, this);
            sas.eventBus.subscribe(WelcomeEvent.EventBusChannelId, WelcomeEvent.AnimationEnd, this._onAnimationEnd, this);

            // add pages
            this._aPages = [];
            PAGE_VIEW_NAMES.forEach(function(sViewName, i) {
                var oPage = sap.ui.jsview(sViewName);
                oPage.setVisible(false);
                oPage.setHeight(null);
                oPage.addStyleClass("sasWelcomeDialogPage");

                this.addContent(oPage);
                this._aPages.push(oPage);
            }.bind(this));

            // open the dialog
            this._oAnimationPromise = null;
            this._fnAnimationPromiseFulfill = null;
            this.open(); // TODO

            // show first page
            var sURIValue = jQuery.sap.getUriParameters().get("sas-welcome"),
                iPage = parseInt(sURIValue);

            this._mStyleClasses = {};

            this._iPageIdx = (!isNaN(iPage) && iPage >= 1 && iPage <= PAGE_VIEW_NAMES.length ? iPage - 1 : 0);
            this._sPageStyleClass = null;

            this.setBusyIndicatorDelay(0);
            this.setBusy(true);
            this._showPage(Animation.Direction.Forward).catch(function(e) {}).then(this.setBusy.bind(this, false));
        },

        setBusy: function(bBusy) {
            this._oLogger.debug("setBusy(" + JSON.stringify(bBusy) + ")");
            Dialog.prototype.setBusy.apply(this, arguments);
        },

        renderer: {},

        onAfterRendering: function() {
            this._oLogger.debug("onAfterRendering()");

            if (Dialog.prototype.onAfterRendering) {
                Dialog.prototype.onAfterRendering.apply(this, arguments);
            }

            var jDiv,
                sStyleClasses = (this._mStyleClasses ? Object.keys(this._mStyleClasses).join(" ") : "");

            if (jQuery("#sas-welcome-background-dots").length === 0) {
                jDiv = jQuery('<div id="sas-welcome-background-dots" class="sasWelcomeBackground sasWelcomeBackgroundDots ' + sStyleClasses + '" ></div>');
                jQuery("#sap-ui-static").append(jDiv);
            }
            if (jQuery("#sas-welcome-background-blue").length === 0) {
                jDiv = jQuery('<div id="sas-welcome-background-blue" class="sasWelcomeBackground sasWelcomeBackgroundBlue ' + sStyleClasses + '" ></div>');
                jQuery("#sap-ui-static").append(jDiv);
            }
        },

        _addStyleClasses: function(vStyleClasses) {
            var sStyleClasses = (jQuery.isArray(vStyleClasses) ? vStyleClasses.join(" ") : vStyleClasses),
                aStyleClasses = (jQuery.isArray(vStyleClasses) ? vStyleClasses : [vStyleClasses]);

            this.addStyleClass(sStyleClasses);
            jQuery(".sasWelcomeBackground").addClass(sStyleClasses);
            aStyleClasses.forEach(function(sStyleClass) {
                this._mStyleClasses[sStyleClass] = true;
            }.bind(this));

            return this;
        },

        _removeStyleClasses: function(vStyleClasses) {
            var sStyleClasses = (jQuery.isArray(vStyleClasses) ? vStyleClasses.join(" ") : vStyleClasses),
                aStyleClasses = (jQuery.isArray(vStyleClasses) ? vStyleClasses : [vStyleClasses]);

            this.removeStyleClass(sStyleClasses);
            jQuery(".sasWelcomeBackground").removeClass(sStyleClasses);
            aStyleClasses.forEach(function(sStyleClass) {
                delete this._mStyleClasses[sStyleClass];
            }.bind(this));

            return this;
        },

        _showPage: function(sAnimationDirection) {
            this._oLogger.debug("_showPage()");

            jQuery.sap.assert(this._aPages.filter(function(oPage) {
                return oPage.getVisible();
            }).length === 0, "hide previous page first");
            jQuery.sap.assert(this._iPageIdx >= 0 && this._iPageIdx < PAGE_VIEW_NAMES.length, "page index out of range");

            this._resetAnimationTimer();

            var oPage = this._aPages[this._iPageIdx];
            this._sPageStyleClass = oPage.getStyleClass();

            this._addStyleClasses([
                this._sPageStyleClass,
                ANIMATION_STYLE_CLASS,
                sAnimationDirection,
                Animation.Action.Enter]);

            oPage.setVisible(true);

            return this._waitForAnimationTimer().then(function() {
                this._removeStyleClasses(ALL_ANIMATION_STYLE_CLASSES);
            }.bind(this));
        },

        _hidePage: function(sAnimationDirection) {
            this._oLogger.debug("_hidePage()");

            var oPage = this._aPages[this._iPageIdx];

            jQuery.sap.assert(oPage.getVisible(), "page already hidden");

            this._resetAnimationTimer();

            this._addStyleClasses([
                ANIMATION_STYLE_CLASS,
                sAnimationDirection,
                Animation.Action.Exit]);

            return this._waitForAnimationTimer().then(function() {
                // hide page, render new DOM, then remove style classes
                oPage.setVisible(false);
                return this._delay(0);
            }.bind(this)).then(function() {
                this._removeStyleClasses(ALL_ANIMATION_STYLE_CLASSES)._removeStyleClasses(this._sPageStyleClass);
                this._sPageStyleClass = null;
            }.bind(this));
        },

        close: function() {
            this._oLogger.debug("close()");

            // save welcome preference to prevent dialog from appearing again for this user
            sap.ui.require([
                "sas/hc/ui/services/petrichor/preferences/PreferencesService"
            ], function(PreferencesService) {
                var oPrefsService = new PreferencesService;
                oPrefsService.setPreference({
                    id: "sas.welcome.suppress",
                    value: "true",
                    application: "sas.welcome"
                });
            });

            jQuery(".sasWelcomeBackground").remove();

            Dialog.prototype.close.apply(this, arguments);
            this.destroy();
        },

        _delay: function(iMs) {
            return new Promise(function(fnFulfill, fnReject) {
                setTimeout(function() {
                    fnFulfill();
                }, (iMs != null ? iMs : 0)); // eslint-disable-line no-eq-null
            });
        },

        _resetAnimationTimer: function() {
            this._oLogger.debug("_resetAnimationTimer()");

            this._oAnimationPromise = new Promise(function(fnFulfill, fnReject) {
                this._fnAnimationPromiseFulfill = fnFulfill;
            }.bind(this));
        },

        _waitForAnimationTimer: function() {
            this._oLogger.debug("_waitForAnimationTimer()");

            // limit animations to MAX_ANIMATION_DURATION_MS
            return Promise.race([
                this._oAnimationPromise,
                this._delay(MAX_ANIMATION_DURATION_MS)
            ]);
        },

        _onPrevious: function(sChannelId, sEventId, oData) {
            this._oLogger.debug("_onPrevious()");

            this.setBusy(true);
            this._hidePage(Animation.Direction.Back).then(function() {
                --this._iPageIdx;
                return this._showPage(Animation.Direction.Back);
            }.bind(this)).catch(function(e) {}).then(this.setBusy.bind(this, false));
        },

        _onNext: function(sChannelId, sEventId, oData) {
            this._oLogger.debug("_onNext()");

            this.setBusy(true);
            this._hidePage(Animation.Direction.Forward).then(function() {
                ++this._iPageIdx;
                if (this._iPageIdx >= PAGE_VIEW_NAMES.length) {
                    this.close();
                    return;
                }
                return this._showPage(Animation.Direction.Forward);
            }.bind(this)).catch(function(e) {}).then(this.setBusy.bind(this, false));
        },

        _onSkip: function(sChannelId, sEventId, oData) {
            this._oLogger.debug("_onSkip()");

            this.setBusy(true);
            this._hidePage(Animation.Direction.Skip).then(function() {
                this.close();
            }.bind(this)).catch(function(e) {}).then(this.setBusy.bind(this, false));
        },

        _onAnimationStart: function(sChannelId, sEventId, oData) {
            this._oLogger.debug("_onAnimationStart()");
        },

        _onAnimationEnd: function(sChannelId, sEventId, oData) {
            this._oLogger.debug("_onAnimationEnd()");

            if (this._fnAnimationPromiseFulfill) {
                this._oLogger.debug("_onAnimationEnd(): _fnAnimationPromiseFulfill()");

                var fnFulfill = this._fnAnimationPromiseFulfill;
                this._fnAnimationPromiseFulfill = null;
                fnFulfill();
            }
        }
    });
});
