//@sasPreCore - Annotation to signal build to merge this file into preCore.js.
sap.ui.define(['jquery.sap.global',
    './ThemeDescriptor',
    'sas/hc/ui/core/util/ComponentServiceClassResolver',
    'sas/hc/ui/core/util/ServicesUtil',
    'sap/ui/core/Core',
    'sas/hc/ui/core/Core'],

    function(jQuery,ThemeDescriptor, ComponentServiceClassResolver, ServicesUtil){
        /**
         * This class provides functions, events and properties related to themes.
         * @name sas.hc.ui.core.theme.ThemeUtil
         * @class
         * @public
         */
        var ThemeUtil = function(){

            "use strict";

            /**
             * The name of the user theme capability.
             * @constant
             * @public
             * @type {string}
             */
            this.CUSTOMIZE_USER_THEME_CAPABILITY = "CustomizeUserThemes";

            /**
             * The current commons version.  Used to determine the correct path for theme content.
             * @constant
             * @public
             * @type {string}
             */
            this.CURRENT_COMMONS_VERSION = ServicesUtil.getServicePlatform() === "WIP"?"900000":"9.0.0";

            /**
             * Message published by loadReportTheme on a successful load.
             * @constant
             * @public
             * @type {string}
             * @default
             */
            this.REPORT_THEME_LOADED = "reportThemeLoaded";

            /**
             * Message published by loadReportTheme when an error occurs.
             * @constant
             * @public
             * @type {string}
             * @default
             */
            this.REPORT_THEME_LOAD_ERROR = "reportThemeLoadError";

            /**
             * Message published by unloadReportTheme.
             * @constant
             * @public
             * @type {string}
             * @default
             */
            this.REPORT_THEME_UNLOADED = "reportThemeUnloaded";

            /**
             * Message published by addReportThemeStyles.
             * @constant
             * @public
             * @type {string}
             * @default
             */
            this.REPORT_THEME_ADDED = "reportThemeAdded";

            /**
             * Message published by removeReportThemeStyles.
             * @constant
             * @public
             * @type {string}
             * @default
             */
            this.REPORT_THEME_REMOVED = "reportThemeRemoved";

            /**
              * Relative path to the report themes.
              * @constant
              * @private
              * @type {string}
              * @default
              */
            this._REPORT_THEME_RELATIVE_PATH = "sas/hc/graph/reportThemes";

            /**
             * EventBus channel for Themes messages.
             * @constant
             * @public
             * @type {string}
             * @default
             */
            this.THEMES = "Themes";

            // report theme name constants

            /**
             * Theme id for the light report theme.
             * @constant
             * @public
             * @type {string}
             * @default
             */
            this.LIGHT = "light";         // REPORT THEME

            /**
             * Theme id for the dark report theme.
             * @constant
             * @public
             * @type {string}
             * @default
             */
            this.DARK = "dark";           // REPORT THEME

            /**
             * Theme id for the highcontrast report theme.
             * @constant
             * @public
             * @type {string}
             * @default
             */
            this.HIGHCONTRAST = "highcontrast";  // REPORT THEME

            /**
             * Theme id for the snow report theme.
             * @constant
             * @public
             * @type {string}
             * @default
             */
            this.SNOW = "snow";  // REPORT THEME

            /**
             * Theme id for the corporate report theme.
             * @constant
             * @public
             * @type {string}
             * @default
             * @deprecated Use SNOW instead.
             */
            this.CORPORATE = this.SNOW;  // REPORT THEME

            /**
             * Theme id for the default report theme.
             * @constant
             * @public
             * @type {string}
             * @default
             */
            this.DEFAULT_REPORT_THEME = this.SNOW; // REPORT THEME

            // application theme name constants

            /**
             * Theme id for the corporate application theme.
             * @constant
             * @public
             * @type {string}
             * @default
             * @deprecated Since version 2.2. Use constants in sas.hc.ui.core.theme.ApplicationTheme instead.
             */
            this.SAS_CORPORATE = "sas_corporate"; // repeated here as an optimization for the ThemeBootTask so the ApplicationTheme dependency chain is not loaded ApplicationTheme.SAS_CORPORATE;  // APP THEME

            // Currently loaded report theme IDs
            var _currentReportThemes = [];

            this.getVersionLibraryThemeVersion = function(libraryName, defaultVersion){
                var libraryVersion = defaultVersion;
                var versionInfo = sap.ui.getVersionInfo();
                if(versionInfo && versionInfo.libraries) {
                    versionInfo.libraries.forEach(function (library) {
                        if (library.name === libraryName) {
                            libraryVersion = library.version;
                        }
                    });
                }

                if(libraryVersion){
                    if (ServicesUtil.getServicePlatform() === "WIP") {
                        if (libraryVersion.indexOf(".") !== -1) {
                            libraryVersion = libraryVersion.substring(0, libraryVersion.indexOf("."));
                        }

                        //replace the last 3 digits since they represent maintenance and hotfixes
                        libraryVersion = libraryVersion.substring(0, libraryVersion.length-3) + "000";
                    }
                    else {
                        if (libraryVersion.indexOf("-") !== -1) {
                            libraryVersion = libraryVersion.substring(0, libraryVersion.indexOf("-"));
                        }

                        if (libraryVersion.indexOf(".") !== -1 && libraryVersion.indexOf(".") !== libraryVersion.lastIndexOf(".")) {
                            libraryVersion = libraryVersion.substring(0, libraryVersion.lastIndexOf("."));
                        }
                    }
                }

                return libraryVersion;
            };

            /**
             * Loads a report theme for a given report theme ID.
             * Publishes ThemeUtil.REPORT_THEME_LOADED on the ThemeUtil.THEMES channel of sas.eventBus for a successful load
             * Publishes ThemeUtil.REPORT_THEME_LOAD_ERROR on the ThemeUtil.THEMES channel of sas.eventBus if an error occurs during loading.
             * @param {string} reportThemeId the theme to load
             * @public
             */
            this.loadReportTheme = function(reportThemeId){
                // Don't attempt to load if theme already loaded
                if (jQuery.inArray(reportThemeId, this.getCurrentReportThemes()) !== -1){
                    sas.eventBus.publish(this.THEMES, this.REPORT_THEME_LOADED, {themeId: reportThemeId});
                    sas.log.info("ThemeUtil.loadReportTheme(): " + reportThemeId + " already loaded.");
                    return;
                }

                var that = this;

                var themeService = ComponentServiceClassResolver.getServiceClassInstance(ComponentServiceClassResolver.THEME_UTIL_SERVICE);
                var repThemeUrl;

                if(themeService){
                    var options = {
                        async: false,
                        id: reportThemeId,
                        dataType: "json",
                        headers: {Accept: "application/vnd.sas.theme.report+json"}
                    };
                    themeService.get(options)
                        .done(function(result){
                            if(result) {
                                if (result.rootUrl) {
                                    repThemeUrl = result.rootUrl + sas.themes.getVersionLibraryThemeVersion("sas.hc.graph", sas.themes.CURRENT_COMMONS_VERSION) + result.htmlStyleSheetPath;
                                } else {
                                    repThemeUrl = sas.config.themeUrl + "/" + that._REPORT_THEME_RELATIVE_PATH + "/" + reportThemeId + "/" + result.htmlStyleSheetName;
                                }
                            }
                        })
                        .fail(function(error){
                            sas.log.error("ThemeUtil.loadReportTheme(): Load failed for report theme " + reportThemeId + ": "+error);
                        });
                }else{
                    var repTheme = sas.themeDescriptor.getReportTheme(reportThemeId);
                    if (repTheme !== null) {
                        var repThemeUrl = sas.config.getThemeUrl();
                        repThemeUrl = repThemeUrl + (repThemeUrl.slice(-1) === "/" ? "" : "/") + this._REPORT_THEME_RELATIVE_PATH + "/" + reportThemeId + "/" + repTheme.stylesheets[0].url;
                    }
                }


                if(repThemeUrl){

                    // Add a link element for this report theme's stylesheet
                    sas.log.trace("ThemeUtil.loadReportTheme(): Calling includeStylesheet()");
                    jQuery.sap.includeStyleSheet(repThemeUrl, reportThemeId + "ReportStylesheet",
                        function(){
                            if (jQuery.inArray(reportThemeId, that.getCurrentReportThemes()) === -1){
                                that.getCurrentReportThemes().push(reportThemeId);
                            }
                            that._publishReportThemeLoaded(reportThemeId);
                        },
                        function(){
                            that._publishReportThemeLoadError(reportThemeId);
                        }
                    );
                } else {
                    sas.eventBus.publish(this.THEMES, this.REPORT_THEME_LOAD_ERROR, {themeId: reportThemeId});
                    sas.log.error("ThemeUtil.loadReportTheme(): Report theme " + reportThemeId + " is not defined.");
                }
            };

            /**
             * Publishes the REPORT_THEME_LOAD_ERROR event.  Used by ThemeUtil and LessController to signal when a report
             * theme has failed to load.
             * @param reportThemeId
             * @private
             */
            this._publishReportThemeLoadError = function(reportThemeId){
                sas.eventBus.publish(this.THEMES, this.REPORT_THEME_LOAD_ERROR, {themeId: reportThemeId});
                sas.log.error("ThemeUtil.loadReportTheme(): Load failed for report theme " + reportThemeId);
            };

            /**
             * Publishes the REPORT_THEME_LOADED event.  Used by ThemeUtil and LessController to signal when a report
             * theme has been successfully loaded.
             * @param reportThemeId
             * @private
             */
            this._publishReportThemeLoaded = function(reportThemeId){
                sas.eventBus.publish(this.THEMES, this.REPORT_THEME_LOADED, {themeId: reportThemeId});
                sas.log.info("ThemeUtil.loadReportTheme(): Report theme " + reportThemeId + " loaded.");
            };

            /**
             * Unloads a report theme for a given report theme ID.
             * Publishes ThemeUtil.REPORT_THEME_UNLOADED on the ThemeUtil.THEMES channel of sas.eventBus for a successful unload
             * @param {string} reportThemeId the report theme to unload
             * @public
             */
            this.unloadReportTheme = function(reportThemeId){
                var i = this.getCurrentReportThemes().indexOf(reportThemeId);
                if (i >= 0){
                    // Remove the link element for this report theme's stylesheet
                    jQuery("#" + reportThemeId + "ReportStylesheet").remove();
                    this.getCurrentReportThemes().splice(i, 1);
                    sas.log.info("ThemeUtil.unloadReportTheme(): Report theme " + reportThemeId + " unloaded.");
                } else {
                    sas.log.info("ThemeUtil.unloadReportTheme(): report theme " + reportThemeId + " not currently loaded.");
                }
                sas.eventBus.publish(this.THEMES, this.REPORT_THEME_UNLOADED, {themeId: reportThemeId});
            };

            /**
             * Returns an array of loaded report theme IDs
             * @returns {string[]} array of report theme IDs
             * @public
             * @public
             */
            this.getCurrentReportThemes = function(){
                return _currentReportThemes;
            };

            /**
             * Adds the report theme ID to the given element's class attribute to allow
             * report theme styles to be applied.
             * Publishes ThemeUtil.REPORT_THEME_ADDED on the ThemeUtil.THEMES channel of sas.eventBus for a successful add.
             * @param {string} reportThemeId the report theme to apply
             * @param {string} elementId the element to modify
             * @public
             */
            this.addReportThemeStyles = function(reportThemeId, elementId){
                if (elementId){
                    var element = jQuery("#" + elementId);
                    if (!element.hasClass(reportThemeId)){
                        element.addClass(reportThemeId);
                    }
                    sas.eventBus.publish(this.THEMES, this.REPORT_THEME_ADDED, {themeId: reportThemeId, elementId: elementId});
                }
            };

            /**
             * Removes the report theme ID from the given element's class attribute.
             * Publishes ThemeUtil.REPORT_THEME_REMOVED on the ThemeUtil.THEMES channel of sas.eventBus for a successful remove.
             * @param {string} reportThemeId the report theme to remove
             * @param {string} elementId the element to modify
             * @public
             */
            this.removeReportThemeStyles = function(reportThemeId, elementId){
                if (elementId){
                    var element = jQuery("#" + elementId);
                    if (element.hasClass(reportThemeId)){
                        element.removeClass(reportThemeId);
                    }
                    sas.eventBus.publish(this.THEMES, this.REPORT_THEME_REMOVED, {themeId: reportThemeId, elementId: elementId});
                }
            };

            /**
             * Returns the base url containing the default graph styles for the current theme. The sas.hc.graph library must be loaded or this method will return <code>null</code>.
             * @public
             * @returns {string} Full path that can be used for loading the graph stylesheet or <code>null</code> if the sas.hc.graph library is not loaded.
             * TODO graph should use a sas.hc.graph library and allow the library to load the styles rather than this mechanism and this method should be deprecated.
             */
            this.getGraphStylesheetUrl = function(){
                return this.getThemeResourceUrl("sas.hc.graph", "graph.css");
            };

            /**
             * Returns the url to the graph stylesheet for the given report theme.
             * @param {string} reportThemeId the report theme to get the stylesheet for
             * @returns {string} Full path that can be used for loading the report theme graph stylesheet
             * @public
             */
            this.getReportGraphStylesheetUrl = function(reportThemeId){

                var themeService = ComponentServiceClassResolver.getServiceClassInstance(ComponentServiceClassResolver.THEME_UTIL_SERVICE);
                var repThemeUrl;
                var that = this;
                if(themeService){
                    var options = {
                        async: false,
                        id: reportThemeId,
                        dataType: "json",
                        headers: {Accept: "application/vnd.sas.theme.report+json"}
                    };
                    themeService.get(options)
                        .done(function(result){
                            if(result) {
                                if (result.rootUrl) {
                                    repThemeUrl = result.rootUrl + sas.themes.getVersionLibraryThemeVersion("sas.hc.graph", sas.themes.CURRENT_COMMONS_VERSION) + result.reportStyleSheetPath;
                                } else {
                                    repThemeUrl = sas.config.themeUrl + "/" + that._REPORT_THEME_RELATIVE_PATH + "/" + reportThemeId + "/" + result.reportStyleSheetName;
                                }
                            }
                        })
                        .fail(function(error){
                            sas.log.error("ThemeUtil.loadReportTheme(): Load failed for report theme " + reportThemeId + ": "+error);
                        });
                }else{
                    repThemeUrl = sas.config.themeUrl + "/" + this._REPORT_THEME_RELATIVE_PATH + "/"+reportThemeId+"/graph.css";
                }

                return repThemeUrl;
            };

            /**
             * Given a relative filename such as MyAudioFile.mp3 compute the full path to
             * where it is located in the theme directory.
             *
             * In the current implementation audio files while stored with themes are not changed per theme but rather a theme independent resource and always loaded from embedded themes.
             *
             * @param {string} sAudioFileName Relative path to the desired file, starting at the audio directory
             * @returns {string} Full path that can be used for loading the audio file
             * @public
             */
            this.getAudioPath = function(sAudioFileName) {
                return sas.config.getThemeUrl() + "/" + sAudioFileName;
            };

            /**
             * Wrapper for private sap.ui.Core _getThemePath method that appends a resource path. This API insulates consumers from changes to the core API.
             * @param {string} library The name of the library to look the theme resource up for. Example sas.hc.ui.commons.
             * @param {string} resourcePath The relative path to the resource to be looked up.
             * @param {string} [theme] The theme to lookup the resource in. Defaults to the currently applied theme if not specified.
             * @returns {string} The url to the resource.
             * @public
             * @see {@link https://openui5.hana.ondemand.com/docs/api/symbols/sap.ui.html#.resource|sap.ui.resource}
             */
            this.getThemeResourceUrl = function(library, resourcePath, theme) {
                var core = sap.ui.getCore(),
                    theme = theme || core.getConfiguration().getTheme(),
                    relativePath;
                // remove leading / if present
                if (resourcePath.charAt(0) === '/') {
                    if (resourcePath.length === 1) {
                        relativePath = "";
                    } else {
                        relativePath = resourcePath.substring(1);
                    }
                } else {
                    relativePath = resourcePath;
                }

                if(ServicesUtil.getServicePlatform() === "WIP" && !relativePath.endsWith(".css")){
                    return sas.config.get("localJSBasePath")+"/"+library.replace(/\./g, "/")+"/themes/sas_corporate/"+relativePath;
                }else {

                    // during library initialization and theme changes, themes for the library are loaded and the theme is
                    // registered as a module using the theme root to redirect to the correct location
                    // see https://openui5.hana.ondemand.com/docs/api/symbols/sap.ui.html
                    relativePath = "themes/" + theme + "/" + relativePath;
                    return sap.ui.resource(library, relativePath);
                }
            };

            /**
             * SAS images used by the icon pool are stored above the themes directory in a library independent "images" directory. Fix up the urls for this fact.
             * @param {string} resourcePath The relative path to the resource to be looked up.
             * @param {string} [theme] The theme to lookup the resource in. Defaults to the currently applied theme if not specified.
             * @param {string} [library] The library this themed image is part of. Defaults to sas.hc.ui.core.
             * @returns {string} The url to the resource.
             * @public
             * @see {@link https://openui5.hana.ondemand.com/docs/api/symbols/sap.ui.html#.resource|sap.ui.resource}
             */
            this.getThemeImageResourceUrl = function(resourcePath, theme, library) {
                if(ServicesUtil.getServicePlatform() === "WIP"){
                    var relativePath;
                    // remove leading / if present
                    if (resourcePath.charAt(0) === '/') {
                        if (resourcePath.length === 1) {
                            relativePath = "";
                        } else {
                            relativePath = resourcePath.substring(1);
                        }
                    } else {
                        relativePath = resourcePath;
                    }
                    return sas.config.get("localJSBasePath")+"/sas/hc/ui/core/themes/sas_base/images/"+relativePath;
                }else {

                    var core = sap.ui.getCore();
                    var theme = theme || core.getConfiguration().getTheme();
                    var library = library || "sas.hc.ui.core";
                    var url = this.getThemeResourceUrl(library, resourcePath, theme);

                    // replace theme name with base/images where images actually live
                    url = url.replace(theme, "sas_base/images");

                    return url;
                }
            };

            this.getCurrentThemeBannerImageUrl = function(callback, currentTheme) {
                if (!currentTheme)
                    currentTheme = sap.ui.getCore().getConfiguration().getTheme();
                var ApplicationTheme = sap.ui.requireSync("sas/hc/ui/core/theme/ApplicationTheme");
                if (currentTheme && currentTheme !== ApplicationTheme.DEFAULT) {
                    var themeService = ComponentServiceClassResolver.getServiceClassInstance(ComponentServiceClassResolver.THEME_UTIL_SERVICE);
                    if (themeService) {
                        var options = {
                            async: false,
                            headers: {
                                "Accept": "application/vnd.sas.theme.application.complete+json"
                            },
                            dataType: "json",
                            id: currentTheme
                        };
                        themeService.get(options).done(function(theme) {
                            if (theme.styleData) {
                                var bannerImageId = theme.styleData["@bannerImageId"];
                                if (bannerImageId) {
                                    callback(ThemeUtil.unquoteImageId(bannerImageId));
                                }
                            }
                        });
                    }
                }
            };

            this.getDefaultApplicationThemeId = function(){
                var themeService = ComponentServiceClassResolver.getServiceClassInstance(ComponentServiceClassResolver.THEME_UTIL_SERVICE);
                var ApplicationTheme = sap.ui.requireSync("sas/hc/ui/core/theme/ApplicationTheme");
                var defaultTheme = ApplicationTheme.DEFAULT;
                if(themeService) {
                    var options = {
                        async: false,
                        dataType: "json",
                        id: "@defaultApplicationTheme"
                    };
                    themeService.get(options).done(function(applicationTheme){
                        defaultTheme = applicationTheme.id;
                    });
                }
                return defaultTheme;
            };

            /**
             * Function that detects the presence of style class mentioned in the function parameter in the control
             * and its parent DOM nodes until the UIArea DOM node using jQuery.closest API.
             *
             * @param {sap.ui.core.Control} control
             * @param {string} styleClass - styleClass name preceded by a dot
             * @return {boolean}
             * @private
             */
            this.isGivenStylePresentInParentChain = function(control, styleClass) {
                var match = false,
                    $control,
                    uiArea,
                    uiAreaDom;

                // check if this is a SAP control
                if(control.getId) {
                    $control = jQuery.sap.byId(control.getId());
                    uiArea = control.getUIArea();
                }
                else {
                    return match;
                }

                if(uiArea)
                    uiAreaDom = jQuery.sap.domById(uiArea.getId());

                //jQuery API - https://api.jquery.com/closest/
                if(!!uiAreaDom && $control.closest(styleClass, uiAreaDom).length === 1) {
                    match = true;
                }

                // If there is no uiAreaDom, maybe because the control is an Icon, check to see if the control itself
                // has the styleClass present.  Removing the '.', if present, because the hasClass method expects the
                // class name without the '.' present
                if ($control && $control.hasClass(styleClass.replace(".", ""))) {
                    match = true;
                }

                return match;
            };

            this.controlShouldInheritStyleClass = function(oControl, sStyleClass) {
                var oControlToCheck = oControl;
                var sStyleClassQuery = "."+sStyleClass;

                while (oControlToCheck) {
                    // While walking up the parent hierarchy not every control will have
                    // the hasStyleClass() or $() method.
                    var bHasStyleClass = jQuery.isFunction(oControlToCheck.hasStyleClass)
                            && oControlToCheck.hasStyleClass(sStyleClass);
                    var bAncestorHasStlyeClass = jQuery.isFunction(oControlToCheck.$)
                            && oControlToCheck.$().closest(sStyleClassQuery).length > 0;

                    if (bHasStyleClass || bAncestorHasStlyeClass) {
                        return true;
                    }

                    // In theory all of the controls in the chain should have a getParent method, even
                    // though the root one will return null from it, but we want to be safe
                    if (jQuery.isFunction(oControlToCheck.getParent)) {
                        oControlToCheck = oControlToCheck.getParent();
                    } else {
                        oControlToCheck = null;
                    }
                }

                return false;
            };
        };

        ThemeUtil.getImageIdFromURL = function(url){
            if(jQuery.sap.startsWith(url,"\"/files/files/")){
                url = url.substring("\"/files/files/".length);
            }
            if(jQuery.sap.endsWith(url,"/content\"")){
                url = url.substring(0, url.length-"/content\"".length);
            }

            return url;
        };

        ThemeUtil.unquoteImageId = function(url){
            if(jQuery.sap.startsWith(url,"\"")){
                url = url.substring(1);
            }
            if(jQuery.sap.endsWith(url,"\"")){
                url = url.substring(0, url.length-1);
            }

            return url;
        };

        ThemeUtil.getURLForImageId = function(imageId){
            return "\"/files/files/"+imageId+"/content\"";
        };

        return ThemeUtil;
    }
,true);
