// (c) 2016, SAS Institute Inc.

sap.ui.define(['jquery.sap.global', 'sap/ui/base/Object', 'sas/hc/ui/core/delegate/Draggable'],
    function(jQuery, BaseObject, Draggable) {
        "use strict";

        var $ = jQuery;

        /**
         * Creates a Draggable delegate that can be attached to Controls requiring
         * capabilities dragging.
         * @class Delegate for dragging Table rows.
         * @author SAS
         * @extends sas.hc.ui.core.delegate.Draggable
         * @version 4.0
         * @constructor
         * @name sas.hc.ui.table.DraggableRowDelegate
         * @public
         */
        var RowDraggable = Draggable.extend("sas.hc.ui.table.DraggableRowDelegate", {

            /**
             * The constructor.
             *
             * @param {sap.ui.core.Control} oControl the Control of which this Draggable is the delegate
             * @param {object} oDragOptions options to utilize while dragging. Based on jQuery UI draggable options.
             * @param {boolean} bEnableTouch set to false to disable touch dragging.
             */

            constructor: function (oControl, oDragOptions, bEnableTouch) {
                //use defaults and allow overrides by consumer
                var oConfig = $.extend({}, {
                    cursor: "move",
                    revert: "invalid",
                    opacity: 0.7,
                    helper: this._fnCreateDragHelper
                }, oDragOptions);

                Draggable.call(this, oControl, oConfig, bEnableTouch);
            },

            _fnCreateDragHelper: function (event) {
                var $draggable = $(this),    //this will be the draggable element
                    oRow = $draggable.data("sasTableRow"),
                    oTable = oRow.getParent(),
                    $helper,
                    $row,   //this will be what's added to the tbody that shows the table cells
                    $thead; //fixed header requires a little work

                if (oTable.getTotalFrozenColumnCount() > 0) {
                    var oDOMFixed = oRow.getDomRef("fixed"),
                        oDOMScroll = oRow.getDomRef(),
                        $rowFixed = $(oDOMFixed).clone(),
                        $rowScroll = $(oDOMScroll).clone(),
                        $fixedTHead;

                    $row = $rowScroll.prepend($rowFixed.children());
                    $fixedTHead = $(oDOMFixed).closest("table").find("thead > tr > th").clone();
                    $thead = $(oDOMScroll).closest("table").find("thead").clone();
                    $thead.find("tr").prepend($fixedTHead);
                } else {
                    $row = $draggable.clone();
                    $thead = $draggable.closest("table").find("thead").clone();
                }

                //create new elements for a faux-table structure to house our cloned row elements
                var $sapUiTable = $("<div></div>").addClass("sapUiTable"),
                    $sapUiTableCnt = $("<div></div>").addClass("sapUiTableCnt"),
                    $sapUiTableCCnt = $("<div></div>").addClass("sapUiTableCCnt"),
                    $sapUiTableCtrlScr = $("<div></div>").addClass("sapUiTableCtrlScr"),
                    $sapUiTableCtrlCnt = $("<div></div>").addClass("sapUiTableCtrlCnt"),
                    $sapUiTableCtrl = $("<table></table>").addClass("sapUiTableCtrl"),
                    $tbody = $("<tbody></tbody>");

                //build the faux-table structure
                $sapUiTable.append($sapUiTableCnt); //outer table element
                $sapUiTableCnt.append($sapUiTableCCnt); //table contents container
                $sapUiTableCCnt.append($sapUiTableCtrlScr); //scrollable container
                $sapUiTableCtrlScr.append($sapUiTableCtrlCnt);  //HTML Table element container
                $sapUiTableCtrlCnt.append($sapUiTableCtrl); //table element
                $sapUiTableCtrl.append($thead); //table is composed of a thead containing TH elements to give widths
                $sapUiTableCtrl.append($tbody); //and tbody which includes all cloned TD elements for the dragged cells
                $tbody.append($row);    //append the actual cloned cell elements
                $sapUiTable.width(oTable.$().width());  //give the outer table element a width to constrain it
                $helper = $sapUiTable;  //and keep the results as the new helper

                //need to put the ghost outside the table somewhere so that it can be dragged around the page without being clipped
                var oOffs = $helper.offset();   //where on the page are we?
                $helper.appendTo($(document.body));
                $helper.offset(oOffs);          //update position to match where we were before

                //S1316617: update cursor position to around middle of ghost
                var oCursorAtOption = { left: Math.round($helper.width()/2), bottom: 5};
                $draggable.data()["ui-draggable"].option("cursorAt", oCursorAtOption);

                return $helper;
            },

            /**
             * Must be set to draggable in onAfterRendering function.
             */
            onAfterRendering: function () {
                this._hasRendered = true;
                var aRows = this._oControl.getRows(),
                    oDragOptions = this._oDragOptions,
                    oConfig,
                    $domRefs;

                //use closure to keep a distinct copy of row and domrefs
                function makeContextChangeCallback(oRow, $domRefs) {
                    return function(oEvent) {
                        if ($domRefs.data("uiDraggable") !== undefined) {
                            $domRefs.draggable("option", 'disabled', (!oRow.getBindingContext()));
                        }
                    };
                }

                aRows.forEach(function(oRow) {
                    //empty/unused rows will have dragging disabled - need to enable drag when populated
                    if (!!oRow._fnDnDDragContextChangeHandler) {
                        //detach existing context change handler if one has already been registered
                        oRow.detachModelContextChange(oRow._fnDnDDragContextChangeHandler);
                    }

                    $domRefs = oRow.getDomRef("fixed");
                    $domRefs = $($domRefs).add(oRow.getDomRef());
                    oConfig = $.extend({}, oDragOptions, {
                        drag: function(event, ui) {
                            ui.droppable = $(this).data("current-droppable");
                            if (ui.droppable) {
                                var feedback = ui.droppable.data("dragFeedback");
                                if (feedback && jQuery.isFunction(feedback.showDragFeedback)) {
                                    feedback.showDragFeedback(event, ui);
                                }
                            }

                            if (jQuery.isFunction(oDragOptions.drag)) {
                                oDragOptions.drag(event, ui);
                            }
                        },
                        disabled: (!oRow.getBindingContext())
                    });
                    $domRefs.draggable(oConfig);
                    $domRefs.data("sasTableRow", oRow);
                    //check for 'group' membership and add data if present
                    if (oDragOptions.hasOwnProperty("group")) {
                        $domRefs.data("sasDnDGroup", oDragOptions.group);
                    }

                    oRow._fnDnDDragContextChangeHandler = makeContextChangeCallback(oRow, $domRefs);
                    oRow.attachModelContextChange(oRow._fnDnDDragContextChangeHandler);
                });
            }

        });
        return RowDraggable;
    },
    true
);
